'''
Handprint: experiment with handwritten text recognition on Caltech Archives.
'''

from .__version__ import __version__, __title__, __description__, __url__
from .__version__ import __author__, __email__
from .__version__ import __license__, __copyright__

# Output format for the files we write.

_OUTPUT_FORMAT = 'png'
_OUTPUT_EXT = '.png'
