# =============================================================================
# @file    __version__.py
# @brief   handprint version info
# @author  Michael Hucka <mhucka@caltech.edu>
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/caltechlibrary/handprint
# =============================================================================

__version__     = '1.0.1'
__title__       = 'handprint'
__description__ = 'Handprint: HANDwritten Page RecognitIoN Test'
__url__         = 'https://github.com/caltechlibrary/handprint'
__author__      = 'Michael Hucka'
__email__       = 'mhucka@library.caltech.edu'
__license__     = 'BSD 3-clause license'
__copyright__   = 'Copyright (C) 2018-2019 by the California Institute of Technology'
